/**
 ** search.c - helper program to search a binary file for a set
 **            sequence of hex values. It will display the
 **            offset along with the matched pattern.
 **
 **     usage: the program takes a pattern of space-separated
 **            byte values (0-9a-f), along with a half-byte
 **            "any value" regex: '.'. Source data is given
 **            via STDIN (typically via the receiving end of
 **            a pipe.
 **
 **     notes: logic isn't thorough; command-line processing
 **            assumes user always does the right thing.
 **
 **            program logic assumes only 1 byte values.
 **
 **   compile: gcc -o search search.c -Wall
 **   execute: cat file | search 'de ad be ef'
 **        or: echo "de ad be ef 00 01 5a" | search 'de ad'
 **/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char      uc;
typedef   signed short int ssi;
typedef unsigned short int usi;
typedef unsigned long  int uli;

int main(int argc, char **argv)
{
    //////////////////////////////////////////////////////////////////
    //
    // A few variables to help along the way
    //
    FILE *input                          = NULL;
    int   index                          = 0;
    int   pos                            = 0;
    uc    upper_nibble                   = 0;
    uc    lower_nibble                   = 0;
    int   value                          = 0;
    uc    length                         = 0;
    uc    regex                          = 0;
    uc    flag                           = 0;
    uli   addr                           = 0;
    uli   matchaddr                      = 0;
    usi  *data                           = NULL;
    uc   *match                          = NULL;

    //////////////////////////////////////////////////////////////////
    //
    // Not enough arguments (or too many) were given
    //
    if ((argc                           <  2) ||
        (argc                           >  3))
    {
        fprintf(stderr, "Error: invalid argument count!\n");
        fprintf(stderr, "Usage: %s PATTERN [FILE]\n", *(argv+0));
        exit(1);
    }

    //////////////////////////////////////////////////////////////////
    //
    // Quick parse of pattern to get rough initial length
    //
    while (*(*(argv+1)+index)           != '\0')
    {
        if (*(*(argv+1)+index)          != ' ')
        {
            length                       = length + 1;
        }
        index                            = index  + 1;
    }

    //////////////////////////////////////////////////////////////////
    //
    // Check integrity of pattern
    //
    if ((length % 2)                    == 1)
    {
        fprintf(stderr, "Error: Malformed PATTERN '%s'\n", *(argv+1));
        exit(1);
    }

    //////////////////////////////////////////////////////////////////
    //
    // Check for optional file argument (read from STDIN otherwise)
    //
    if (argc                            == 2)
    {
        input                            = stdin;
    }
    else
    {
        input                            = fopen(*(argv+2), "r");
        if (input                       == NULL)
        {
            fprintf(stderr, "Error: Could not open '%s' for reading!\n",
                            *(argv+2));
            exit(1);
        }
    }

    /////////////////////////////////////////////////////////////////
    //
    // Since we operate in groups of 2 (for a whole byte), cut our
    // length value in half
    //
    length                               = length / 2;

    /////////////////////////////////////////////////////////////////
    //
    // Allocate storage for our data and matched patterns
    //
    data                                 = (usi *)malloc(sizeof(usi)*length);
    match                                = (uc *)malloc(sizeof(uc)*length);

    /////////////////////////////////////////////////////////////////
    //
    // Parse pattern for individual values, and construct values
    //
    index                                = 0;
    for (pos = 0; pos < strlen(*(argv+1)); pos += 2)
    {
        /////////////////////////////////////////////////////////////
        //
        // regex is our regular expression symbol detection flag
        //
        regex                            = 0;

        /////////////////////////////////////////////////////////////
        //
        // upper_nibble stores the upper 4 bits of the byte
        //
        upper_nibble                     = *(*(argv+1)+pos);

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is a number, adjust accordingly
        //
        if ((upper_nibble               >= '0') &&
            (upper_nibble               <= '9'))
        {
            upper_nibble                 = upper_nibble - 48;
        }

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is a lowercase letter
        //
        else if ((upper_nibble          >= 'a') &&
                 (upper_nibble          <= 'f'))
        {
            upper_nibble                 = upper_nibble - 87;
        }

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is an uppercase letter
        //
        else if ((upper_nibble          >= 'A') &&
                 (upper_nibble          <= 'F'))
        {
            upper_nibble                 = upper_nibble - 55;
        }

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is a regular expression
        //
        else if (upper_nibble           == '.')
        {
            regex                        = 1;
            upper_nibble                 = 0;
        }

        /////////////////////////////////////////////////////////////
        //
        // If none of these, we have an error
        //
        else
        {
            fprintf(stderr, "Error: Invalid pattern symbol '%c' (%hhu)\n",
                            upper_nibble,
                            upper_nibble);
            exit(1);
        }

        /////////////////////////////////////////////////////////////
        //
        // Position upper_nibble to the upper-order bits of byte,
        // and scrub any errant values not in those bit positions
        //
        upper_nibble                     = upper_nibble << 4;
        upper_nibble                     = upper_nibble &  0x00F0;

        /////////////////////////////////////////////////////////////
        //
        // lower_nibble stores the lower 4 bits of the byte
        //
        lower_nibble                     = *(*(argv+1)+(pos+1));

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is a number, adjust accordingly
        //
        if ((lower_nibble               >= '0') &&
            (lower_nibble               <= '9'))
        {
            lower_nibble                 = lower_nibble - 48;
        }

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is a lowercase letter
        //
        else if ((lower_nibble          >= 'a') &&
                 (lower_nibble          <= 'f'))
        {
            lower_nibble                 = lower_nibble - 87;
        }

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is an uppercase letter
        //
        else if ((lower_nibble          >= 'A') &&
                 (lower_nibble          <= 'F'))
        {
            lower_nibble                 = lower_nibble - 55;
        }

        /////////////////////////////////////////////////////////////
        //
        // Check to see if it is a regular expression
        //
        else if (lower_nibble           == '.')
        {
            regex                        = regex | 2;
            lower_nibble                 = 0;
        }

        /////////////////////////////////////////////////////////////
        //
        // If none of these, we have an error
        //
        else
        {
            fprintf(stderr, "Error: Invalid pattern symbol '%c' (%hhu)\n",
                            lower_nibble,
                            lower_nibble);
            exit(1);
        }

        /////////////////////////////////////////////////////////////
        //
        // Scrub any errant values not in those bit positions
        //
        lower_nibble                     = lower_nibble &  0x000F;

        /////////////////////////////////////////////////////////////
        //
        // Pack the values into one element in data array, based on
        // whether or not there is a regex, and where it is
        //
        *(data+index)                    = 0;
        if (regex                       == 0)
        {
            *(data+index)                = (upper_nibble | lower_nibble) & 0x00FF;
        }
        else if (regex                  == 1)
        {
            *(data+index)                = upper_nibble | lower_nibble   | 0xF000;
        }
        else if (regex                  == 2)
        {
            *(data+index)                = upper_nibble | lower_nibble   | 0x0F00;
        }
        else
        {
            *(data+index)                = upper_nibble | lower_nibble   | 0xFF00;
        }

        /////////////////////////////////////////////////////////////
        //
        // Increment index
        //
        index                            = index  + 1;

        /////////////////////////////////////////////////////////////
        //
        // Support space-less input (experimental)
        //
        if (*(*(argv+1)+(pos+2))        == 0x20)
            pos                          = pos    + 1;
    }

    /////////////////////////////////////////////////////////////////
    //
    // Read through the input data looking for set patterns
    //
    pos                                  = 0;
    while (!feof(input))
    {
        /////////////////////////////////////////////////////////////
        //
        // Obtain a byte from the file (STDIN or specified file)
        //
        value                            = fgetc(input);

        /////////////////////////////////////////////////////////////
        //
        // If we have an exact match, record it in our match array
        //
        if (value                       == *(data+pos))
        {
            *(match+pos)                 = value  & 0xFF;
            pos                          = pos    + 1;
            flag                         = flag   + 1;
        }

        /////////////////////////////////////////////////////////////
        //
        // If we have an upper nibble regex, lower nibble number
        // match, record it in our match array
        //
        else if ((*(data+pos) & 0xFF00) == 0xF000)
        {
            /////////////////////////////////////////////////////////
            //
            // Verify the lower nibble value
            //
            if ((*(data+pos)  & 0x000F) == (value & 0x000F))
            {
                *(match+pos)             = value  & 0xFF;
                pos                      = pos    + 1;
                flag                     = flag   + 1;
            }

            /////////////////////////////////////////////////////////
            //
            // If lower nibble doesn't match, we do not have a match,
            // kick out and start looking for the pattern anew
            //
            else
            {
                pos                      = 0;
                flag                     = 0;
            }
        }

        /////////////////////////////////////////////////////////////
        //
        // If we have an lower nibble regex, upper nibble number
        // match, record it in our match array
        //
        else if ((*(data+pos) & 0xFF00) == 0x0F00)
        {
            /////////////////////////////////////////////////////////
            //
            // Verify the upper nibble value
            //
            if ((*(data+pos)  & 0x00F0) == (value & 0x00F0))
            {
                *(match+pos)             = value  & 0xFF;
                pos                      = pos    + 1;
                flag                     = flag   + 1;
            }

            /////////////////////////////////////////////////////////
            //
            // If upper nibble doesn't match, we do not have a match,
            // kick out and start looking for the pattern anew
            //
            else
            {
                pos                      = 0;
                flag                     = 0;
            }
        }

        /////////////////////////////////////////////////////////////
        //
        // If we have an lower nibble regex, upper nibble regex
        // match, record it in our match array
        //
        else if ((*(data+pos) & 0xFF00) == 0xFF00)
        {
            *(match+pos)                 = value  & 0xFF;
            pos                          = pos    + 1;
            flag                         = flag   + 1;
        }

        /////////////////////////////////////////////////////////////
        //
        // We do not have a match, start search anew
        //
        else
        {
            pos                          = 0;
            flag                         = 0;
        }

        /////////////////////////////////////////////////////////////
        //
        // If we have a match (as marked by flag variable), record
        // the matching address in matchaddr
        //
        if (flag                        == 1)
        {
            matchaddr                    = addr;
        }

        /////////////////////////////////////////////////////////////
        //
        // If we've hit our pattern length, AND flag is not 0,
        // display the address and the contents of the matched
        // pattern from the data stream
        //
        if ((pos                         == length) &&
            (flag                        != 0))
        {
            fprintf(stdout, "%.4lx:", matchaddr);
            for (index = 0; index < length; index++)
                fprintf(stdout, "%.2hhx ", (*(match+index) & 0x00FF));
            fprintf(stdout, "\n");
            pos                          = 0;
            matchaddr                    = 0;
        }

        /////////////////////////////////////////////////////////////
        //
        // Increment our offset by one
        //
        addr                             = addr   + 1;
    }

    //////////////////////////////////////////////////////////////////
    //
    // Close any open files, free any allocated memory
    //
    free(data);
    free(match);
    if (input                           != stdin)
        fclose(input);

    return(0);
}
